#define SERVER_NAME "valkey"
#define VALKEY_VERSION "7.2.13"
#define VALKEY_VERSION_NUM 0x0007020D

/* Redis compatibility version, should never
 * exceed 7.2.x. */
#define REDIS_VERSION "7.2.4"
#define REDIS_VERSION_NUM 0x00070204
